from .smmcalibration import AV
from .smmcalibration import PSD
from .smmcalibration import load_trajectory
from .simulations import simulate_trace
from .simulations import downsampled_trace


import os

from pkg_resources import DistributionNotFound,get_distribution

try:
    __version__ = get_distribution("tweezepy").version
except DistributionNotFound:
    __version__ = "unknown version"

#this_directory = os.path.abspath(os.path.dirname(__file__))
#pkginfo_path = os.path.join(this_directory,
#                            'tweezepy_info.json')
#pkginfo = json.load(open(pkginfo_path))

#__version__ = pkginfo["version"]
__all__ = [
           "AV",
           "PSD",
           "simulate_trace",
           "downsampled_trace",
           "load_trajectory",
           "__version__",
           ]


# -*- coding: utf-8 -*-
"""

The smmcalibration module.

Example usage:
    from tweezepy.simulations import simulate_trace
    from tweezepy.smmcalibration import AV
    trace = simulate_trace()
    av = AV(trace,100)
    av.plot()
    
Created on Mon May  4 11:51:23 2020

Author: Ian L. Morgan
email: ilmorgan@ucsb.edu
"""
import numpy as np
#import pkg_resources

from scipy.signal import welch
from tweezepy.allanvar import avar,totvar
from tweezepy.MLE import MLEfit, Gamma_Distribution
from tweezepy.expressions import aliasPSD,lansdorpPSD,SMMAV

def load_trajectory():
    """
    Loads test bead trajectory sampled at 400 Hz.

    Returns
    -------
    data : array-like
        Bead trajectory data in nm
    """
    import os 
    this_dir, this_filename = os.path.split(__file__)
    fname = os.path.join(this_dir, "data", "trajectory.csv")
    #fname = pkg_resources.resource_stream(__name__, 'data/trajectory.csv')
    data = np.loadtxt(fname, delimiter=',')
    return data

class calibration(MLEfit):
    """
    Base class for PSD and AV calibration.
    
    Parameters
    ----------
    trace : array
        1-D array of bead positions in nm
    fsample : float
        Sampling frequency in Hz

    Attributes
    ----------
    trace : array
        1-D array of bead positions in nm
    fsample : float
        Sampling frequency in Hz
    """
    def __init__(self,trace,fsample):
        self.trace = trace
        self.fsample = fsample
    def plot(self, 
             fig = None,
             fig_kwgs={},
             ax_fit_kwgs = {},
             ax_res_kwgs = {},
             data_label = None,
             fit_label = None,
             data_color = None,
             fit_color = 'k'):
        """
        Utility function for plotting.

        Parameters
        ----------
        fig : Figure, optional
            Matplotlibe Figure object, by default None
        fig_kwgs : dict, optional
            Figure keyword arguments, by default {}
        ax_fit_kwgs : dict, optional
            Axis keyword arguments, by default {}
        ax_res_kwgs : dict, optional
            Axis keyword arguments, by default {}
        data_label : str, optional
            Legend label for data.
        fit_label : str, optional
            Legend label for fit line.

        Returns
        -------
        fig, ax : Figure, Axes
            Figure and axes objects.
        """
        try:
            import matplotlib.pyplot as plt
        except ImportError:
            raise RuntimeError("Matplotlib is required for plotting.")
        data = self.data
        if not isinstance(fig, plt.Figure):
            fig = plt.figure(figsize = (5,5),**fig_kwgs)
        ax = fig.get_axes()
        gs = plt.GridSpec(nrows=2, ncols=1, height_ratios=[4, 1],hspace = 0)
        if len(ax) == 0:
            ax.append(fig.add_subplot(gs[0], **ax_fit_kwgs))    
            if 'yfit' in data.keys():
                ax.append(fig.add_subplot(gs[1], **ax_res_kwgs))
                ax[1].axhline(0,c='k',**ax_res_kwgs)
                ax[1].set_ylabel(r'$\Delta$')
        # Plot data with errorbars    
        eb = ax[0].errorbar(data['x'], data['y'], yerr=data['yerr'],
                            fmt = 'o', zorder = 0, c = data_color,
                            ms=3, capsize = 3, label = data_label)
        ax[0].set_xscale('log')
        ax[0].set_yscale('log')
        # if there is a fit plot it
        if ('yfit' in data.keys()):
            # Plot fit
            c = eb[0].get_color() # get last color
            ax[0].plot(data['x'], data['yfit'],
                       lw = 2, label=fit_label,c=fit_color, zorder = 1) 
            if (len(ax) == 2):
                ax[1].plot(data['x'],data['residuals'])
                ax[1].set_xscale('log')
                ax[0].get_xaxis().set_ticklabels([])
        if isinstance(self,AV):
            ax[-1].set_xlabel(r'$\tau$ (s)')
            ax[0].set_ylabel(r'$\sigma_{AV}^2$ (nm$^2$)')
        if isinstance(self,PSD):
            ax[-1].set_xlabel(r'$f$ (Hz)')
            ax[0].set_ylabel(r'PSD (nm$^2$/Hz)')   
        return fig, ax
    def _make_guess(self, 
                    kT = 4.1,
                    viscosity = 8.94e-10,
                    radius = 530):
        """
        Estimate gamma based on Stokes drag and kappa based on equipartition theorem.

        Parameters
        ----------
        trace : np.array
            Bead positions in nm
        kT : float, optional
            Thermal energy in pN nm, by default 4.1
        viscosity : float, optional
            Dynamic viscosity of solution in pN s/nm^2, by default 8.94e-10
        radius : float, optional
            Bead radius in nm, by default 530

        Returns
        -------
        guess : list
            gamma and kappa guesses
        """
        # Make initial guess for kappa based on equipartition theorem
        kappa = kT/np.var(self.trace)
        # Make initial guess for alpha based on myone bead in water 
        gamma = 6 * np.pi * viscosity * radius 
        guess = [gamma,kappa]    
        return guess
    def _predefined(self,
                    func,
                    gamma = None,
                    kappa = None,
                    tracking_error = False,
                    epsilon = None,
                    ):
        if gamma and kappa and tracking_error:
             func2 = lambda e: func(gamma, kappa, e)
        elif gamma and tracking_error:
            func2 = lambda k,e: func(gamma, k, e)
        elif kappa and tracking_error:
            func2 = lambda g,e: func(g, kappa, e)
        elif epsilon and tracking_error:
            func2 = lambda g,k: func(g, k, epsilon)
        elif gamma:
            func2 = lambda k: func(gamma,k,0)
        elif kappa:
            func2 = lambda g: func(g,kappa,0)
        elif tracking_error:
            func2 = lambda g,k,e: func(g,k,e)
        else:
            func2 = lambda g,k: func(g,k,0)
        return func2
    def mlefit(self,
               fitfunc = None, 
               cutoffs = [-np.inf,np.inf],
               tracking_error = False, 
               guess = None, 
               gamma = None, 
               kappa = None, 
               epsilon = None,
               kT = 4.1, 
               viscosity = 8.94e-10, 
               radius = 530.,  
               pedantic = True, 
               scale_covar = False,
               minimizer_kwargs = {}):
        """
        Estimate parameters and uncertainties via maximum likelihood estimation.

        Parameters
        ----------
        fitfunc : str or function, optional
            String or user defined function, 
            by default 'lansdorpPSD' for PSD and 'SMMAV' for AV
        tracking_error : bool, optional
            Incorporate tracking errors, by default False
        guess : list, optional
            Initial parameter guesses, by default None
        gamma : float, optional
            Fixed value for gamma parameter, by default None
        kappa : float, optional
            Fixed value for kappa parameter, by default None
        epsilon : float, optional
            Fixed value for epsilon parameter, by default None
        pedantic : bool, optional
            Ignore unhelpful warning messages, by default True
        kT : float, optional
            Thermal energy in pN nm, by default 4.1
        viscosity : float, optional
            Dynamic viscosity for initial parameter guesses in pN s/nm^2, by default 8.94e-10
        radius : float, optional
            Bead radius for inital parameter guesses in nm, by default 530.
        minimizer_kwargs : dict, optional
            Keyword arguments to send to scipy.optimize.minimize.
        """
        if cutoffs:
            assert len(cutoffs) == 2, "cutoffs should have a lower an upper cutoff in the form of [lower,upper]."
        msk = (self.data_init['x']>=cutoffs[0])&(self.data_init['x']<=cutoffs[1])
        for key,d in self.data_init.items():
            self.data[key] = d[msk]        
        x = self.data['x']

        if isinstance(self,AV):
            if not fitfunc:
                fitfunc = 'SMMAV'
            if fitfunc == 'SMMAV':
                ts = 1/self.fsample
                func = lambda g,k,e: SMMAV(x,ts,g,k,e,kT=kT)
                func = self._predefined(func,
                                        gamma=gamma,
                                        kappa = kappa,
                                        tracking_error=tracking_error,
                                        epsilon = epsilon,
                                        )
            else:
                func = fitfunc
                if not guess:
                    raise RuntimeError("User-provided function requires initial parameter guesses.")
            
        if isinstance(self,PSD):
            # Select appropriate function
            if not fitfunc:
                fitfunc = 'lansdorpPSD'

            if fitfunc == 'lansdorpPSD':
                func = lambda g,k,e: lansdorpPSD(x,self.fsample,g,k,e,kT=kT)
                func = self._predefined(func,
                                        gamma=gamma,
                                        kappa = kappa,
                                        tracking_error=tracking_error,
                                        epsilon = epsilon,
                                        )
            elif fitfunc == 'aliasPSD':
                func = lambda g,k,e: aliasPSD(x,self.fsample,g,k,e,kT=kT)
                func = self._predefined(func,
                                        gamma=gamma,
                                        kappa = kappa,
                                        tracking_error=tracking_error,
                                        epsilon = epsilon,
                                        )
            else:
                func = fitfunc
                if not guess:
                    raise RuntimeError("User-provided function requires initial parameter guesses.")
        self.func = func
        # If no guess provided, make default guess
        if not guess:
            #make default guess based on Stoke's drag and equipartition theorem
            guess = self._make_guess(kT=kT,
                                     viscosity=viscosity,
                                     radius=radius)
            if gamma:
                guess.pop(0)
            if kappa:
                guess.pop(1)
            if tracking_error:
                guess.append(10)
        self.guess = guess
        MLEfit.__init__(self, pedantic = pedantic, scale_covar = scale_covar, **minimizer_kwargs)
    @property
    def data(self):
        """
        Returns dictionary of data.

        Returns
        -------
        dict
            Dictionary of data.
        """
        return self._data
    @data.setter
    def data(self, value):
        self._data = value
class AV(calibration):
    """
    A class for computing and fitting the Allan variance using maximum likelihood estimation.

    Parameters
    ----------
    trace : array
        Bead trajectory in nm.
    fsample : float
        Sampling frequency in Hz.
    taus : str, optional
        Tau interval sampling. Options are 'octave', 'decade', and 'all', by default 'octave'
    mode : str, optional
        Allan variance type, either 'avar', 'oavar', and 'totvar', by default 'oavar'. 
        See allanvar module for more information.
    edf : str, optional
        Equivalent degrees of freedom for AV. Options are 'real' and 'approx', by default 'real'

    Raises
    ------
    AssertionError
        If taus, mode, or edf is not recognized.

    Examples
    --------
    >>> from tweezepy import load_trajectory, AV
    >>> trace = load_trajectory() # Load trajectory in nm
    >>> av = AV(trace, fsample = 400) # Compute overlapping AV
    >>> av.mlefit() # Perform MLE fit
    >>> print(av.results)
    """

    def __init__(self, trace, fsample,taus = 'octave',mode = 'oavar',edf='real'):
        calibration.__init__(self,trace,fsample)
        assert taus in ['all','octave','decade'], "taus must be either all, octave, or decade."
        assert mode in ['avar','oavar','totvar'], "mode must be either avar, oavar, or totvar."
        assert edf in ['real','approx'], "edf must be either real or approx."
        if mode == 'avar':
            taus, edfs, oavs = avar(trace, rate = fsample, taus = taus,edf=edf, overlapping = False)
        elif mode == 'oavar':
            taus, edfs, oavs = avar(trace, rate = fsample, taus = taus,edf=edf, overlapping = True)
        elif mode == 'totvar':
            taus, edfs, oavs = totvar(trace, rate = fsample, taus = taus,edf=edf)
        self.x = taus
        self.y = oavs
        self.shape = edfs/2
        # Probability distribution function
        self.gd = Gamma_Distribution(self.shape,self.y)
        self.yerr = self.gd.std(self.y)
        self._data = {'x':self.x,
                      'shape':self.shape,
                      'y':self.y,
                      'yerr':self.yerr}
        self.data_init = self.data.copy()
        
class PSD(calibration,MLEfit):
    """
    A class for computing and fitting the power spectral density using maximum likelihood estimation.

    Parameters
    ----------
    trace : array
        Bead positions in nm.
    fsample : float
        Sampling frequency in Hz
    bins : int, optional
        Number of bins, by default 3

    Example
    -------
    >>> from tweezepy import load_trajectory, PSD            
    >>> trace = load_trajectory() # Load trajectory in nm
    >>> psd = PSD(trace, fsample = 400) # Compute PSD using Welch's method
    >>> psd.mlefit() # Perform MLE fit
    >>> print(psd.results)
    """

    def __init__(self, 
                 trace, # input bead trajectory 
                 fsample, # input sampling frequency in Hz
                 bins=3, # input number of half-overlapping bins
                 ):
        calibration.__init__(self, trace, fsample)
        N = len(trace)
        nperseg = (2.*N)/(bins+1)
        f, dens = welch(trace, fsample, nperseg=nperseg,return_onesided=False)
        msk = f>0
        f, dens = f[msk], dens[msk]
        shapes = np.full_like(f,bins)
        self.x = f
        self.shape = shapes
        self.y = dens
        # Probability distribution function
        self.gd = Gamma_Distribution(self.shape,self.y)
        self.yerr = self.gd.std(self.y)
        self._data = {'x':self.x,
                      'shape':self.shape,
                      'y':self.y,
                      'yerr':self.yerr}
        self.data_init = self.data.copy()



"""
This module include functions for running bead trajectory simulations.

"""
import numpy as np
from numba import jit

@jit(nopython=True) # speeds up certain types of code
def simulate_trace(gamma = 1.e-5,
                   kappa = None,
                   fsim = 400,
                   sim_points = 10240,
                   seed = None):
    """Takes parameters and gives a simulated bead trajectory.

    .. math::
        \Delta x = \\frac{\Delta t}{\\alpha}\left(\sqrt{\\frac{2\\alpha k_BT}{dt}} F_L-\\kappa x_{i-1}\\right)
    .. math:: 
        x_i = x_{i-1}+dx

    Parameters
    ----------
    gamma : float, optional
        Stokes dissipation in pNs/nm^2, by default 1.e-5
    kappa : float, optional
        Spring constant in pN/nm, by default None
    fsim : int, optional
        Simulation frequency in Hz, by default 100
    sim_points : int, optional
        Number of bead positions, by default 10240
    seed : int, optional
        Random seed, by default None
    
    Returns
    -------
    xtrace : array
        Array of bead positions.

    """
    if seed != None:
        np.random.seed(seed)
    if not kappa:
        kappa = 2*np.pi*10*gamma
    if fsim < kappa/gamma: # Simulation is unstable below this condition
        raise ValueError(r'The simulation frequency must be larger than kappa/gamma.')
    dt = 1/fsim # time step between collecting probe position
    kT = 4.1    # thermal energy, in pN nm (approx. 4.1 at room temp.)
    std = np.sqrt(2*gamma*kT/dt) # standard deviation of random gaussian force
    F_L = np.random.normal(0,std,sim_points-1) # Random Langevin force
    xtrace = np.zeros(sim_points) 
    for i in range(1,sim_points):
        dx =  dt/gamma * (F_L[i-1]-kappa*xtrace[i-1]) # Smoluchowski eqn
        xtrace[i] = xtrace[i-1] + dx
    return xtrace

def downsampled_trace(gamma = 1.e-5,
                      kappa = None,
                      fsample = 100,
                      N = 10240, 
                      seed = None):
    """
    Simulates and downsamples trace.

    Parameters
    ----------
    gamma : float, optional
        Stokes dissipation in pNs/nm^2, by default 1.e-5
    kappa : float, optional
        Spring constant in pN/nm, by default None
    fsample : int, optional
        Sampling frequency in Hz, by default 100
    N : int, optional
        Number of downsampled bead positions, by default 10240
    seed : int, optional
        Random seed, by default None

    Returns
    -------
    xtrace_ds : numpy.array
        Array of bead positions.

    """
    bin_size = 1000 # points to average when downsampling
    fsim = int(bin_size*fsample) # simulation frequency
    sim_points = N*bin_size # number of points to simulate
    xtrace = simulate_trace(gamma,kappa,fsim,sim_points,seed) # do simulation
    bins = xtrace.reshape((N,bin_size)) # reshape into bins 
    xtrace_ds = np.mean(bins,axis=1) # average bins to produce downsampled trace
    return xtrace_ds


"""
This file is part of tweezepy.py
"""

from autograd import hessian
from autograd.scipy.stats import gamma
from inspect import signature
from scipy.optimize import minimize
from scipy import stats

import autograd.numpy as np

class MCMC:
    """
    Monte Carlo sampler class.

    Parameters
    ----------
    walkers : int, optional
        Number of walkers, by default 32
    steps : int, optional
        Number of steps, by default 1600
    progress : bool, optional
        Print progress bar, by default True
    """
    def __init__(self, walkers = 32, steps = 1600, progress = True,**kwargs):
        try:
            import emcee
        except ImportError:
            RuntimeError("Monte Carlo sampling requires the emcee package.")
        self.walkers = walkers
        self.steps = steps

        scale = np.power(10,np.floor(np.log10(self.params)))
        pos = self.params + 1e-4 * np.random.randn(walkers,self.nparams) * scale
        nwalkers,ndims = pos.shape
        self.sampler = emcee.EnsembleSampler(nwalkers,ndims,self.logL,**kwargs)
        self.sampler.run_mcmc(pos, steps,progress = progress)
        self.samples = self.sampler.get_chain()
        self.autocorr_time = self.sampler.get_autocorr_time()

    def sample_plot(self,
                    fig=None,
                    labels = [],
                    fig_kwgs={},
                    ax_kwgs={}):
        """
        Plot the accepted Monte Carlo samples.

        Parameters
        ----------
        fig : object, optional
            Figure object, by default None
        labels : list, optional
            Plot labels, by default []
        fig_kwgs : dict, optional
            Figure keywords, by default {}
        ax_kwgs : dict, optional
            Axes keywords, by default {}

        Returns
        -------
        fig : Figure
            Figure object.
        axes : Axes
            Axes object.
        """
        try:
            import matplotlib.pyplot as plt
        except ImportError:
            raise RuntimeError("Matplotlib is required for plotting.")
        if not isinstance(fig, plt.Figure):
            fig = plt.figure(figsize = (10,3*self.nparams),**fig_kwgs)
        axes = fig.get_axes()
        gs = plt.GridSpec(nrows=self.nparams,ncols =1)
        if len(labels) == 0:
            labels = self.names
        if len(axes) == 0:
            for i in range(self.nparams):
                axes.append(fig.add_subplot(gs[i]))
        for i,p in enumerate(labels):
            ax = axes[i]
            ax.plot(self.samples[:, :, i], c="k", alpha=0.3,**ax_kwgs)
            ax.axvline(2*self.autocorr_time[i],c="k",lw=2)
            ax.set_xlim(0, len(self.samples))
            ax.set_ylabel(p)
            ax.yaxis.set_label_coords(-0.1, 0.5)
            ax.ticklabel_format(axis='y',style='sci',scilimits=(0,0))
        axes[-1].set_xlabel("step number")
        return fig, axes

    def calc_mc_errors(self, percentiles = [15.87,50,84.13], discard = None, thin = None):
        """
        Computes percentiles from Monte Carlo samples.

        Parameters
        ----------
        percentiles : list, optional
            Percentiles for each parameter, by default [15.87,50,84.13]
        discard : int, optional
            Number of "burn-in" steps to discard, by default 100
        thin : int, optional
            N, by default 10
        
        Return
        ------
        errors : array
            Errors from Monte Carlo sampling.
        """
        tau = max(self.autocorr_time)

        if not discard:
            discard = np.ceil(2*tau)
        if not thin:
            thin = np.ceil(tau/2)
        if discard < 2*tau:
            raise RuntimeError('discard should be greater than twice the autocorrelation time %s.'%self.autocorr_time)
        if thin < tau%2:
            raise RuntimeError('thin should be greater than half the autocorrelation time %s.'%self.autocorr_time)
        self.flat_samples = self.sampler.get_chain(discard=discard,thin = thin,flat = True)
        return np.percentile(self.flat_samples,percentiles,axis=0).T

    def corner_plot(self,
                    quantiles = [0.16,0.84],
                    labels = None,
                    **kwargs):
        """
        Utility function for generating corner plots.

        Parameters
        ----------
        quantiles : list, optional
            Quantiles to annotate, by default (0.16,0.84)
        labels : list, optional
            Parameter labels, by default None

        Returns
        -------
        fig : Figure
            Figure object.
        axes : Axes
            Axes object.
        """
        try:
            import corner
        except ImportError:
            raise RuntimeError("Corner plot requires the corner module.")
        if not labels:
            labels = self.names
        fig = corner.corner(self.flat_samples,  
                            truths=self.params,
                            quantiles=quantiles,
                            #levels=(1-np.exp(-0.5),),
                            #title_fmt = '.2e',
                            #show_titles = True,
                            labels = labels,
                            title_kwargs={'fontdict':{'fontsize':12}},
                            **kwargs)
        ax = fig.get_axes()
        for a in ax:
            if a.get_xlabel():
                a.ticklabel_format(axis='x',style='sci',scilimits=(0,0),useMathText=True)
                t = a.get_xaxis().get_offset_text()
                t.set_position((1.1,.25))
                #t.set_xposition((-.5,0.9))
                #a.ticklabel_format()
            if a.get_ylabel():
                a.ticklabel_format(axis='y',style='sci',scilimits=(0,0),useMathText=True)
                t = a.get_yaxis().get_offset_text()
                t.set_position((-.3,0.9))
        return fig,ax
        
class MLEfit(MCMC):
    """
    Perform maximum likelihood estimation and uncertainty calculations.

    Parameters
    ----------
    pedantic : bool, optional
        Ignore unhelpful warnings,
        by default True
    scale_covar : bool, optional
        Whether to scale standard errors by reduced chi-squared,
        by default False
    fit_kwargs : dict, optional
        Disctionary of keyword arguments passed to scipy.optimize.minimize, 
        by default {}.
    
    Attributes
    ----------
    names : list
        Fit function parameter names.
    params : array
        Parameter values.
    std_errors : array
        Parameter uncertainties.
    chi2 : float
        Chi-squared value.
    redchi2 : float
        Reduced chi-squared value.
    AIC : float 
        Akaike information criterion. 
    AICc : float 
        Corrected Akaike information criterion. 
    """
    def __init__(self, pedantic = True, scale_covar = False, minimizer_kwargs = {}):
        if pedantic == False:
            np.seterr('warn')
        elif pedantic == True:
            np.seterr('ignore')
        # Fancy way of determining fit param names        
        names = signature(self.func).parameters # inspect fit function parameters
        self.names = list(names.keys()) # make list of parameter names
        # Data
        shape = self.data['shape']
        y = self.data['y']
        yerr = self.data['yerr']
        self.ndata = len(y)
        # Log likelihood
        self.gd = Gamma_Distribution(shape,y)
        self.logL = lambda p: self.gd.logpdf(self.func(*p)).sum()
        # Negative log likelihood
        self.negLL =  lambda p: -self.logL(p)
        # Use automatic differentiation to calculate hessian
        hess = hessian(self.negLL)
        # Minimize negative log likelihood
        self.fit = minimize(self.negLL,x0=self.guess,method = 'Nelder-Mead',**minimizer_kwargs)
        # Save minimizer fit results
        self.params = self.fit['x']
        self.nparams = len(self.fit['x'])
        self.success = self.fit['success']
        # Collect results into dictionary
        results = {}
        # Throw warning if fit fails
        if not self.success:
            print('MLE fitting failed to converge. %s'%self.fit['message'])
            #self.params = np.array([float('nan') for i in range(self.nparams)])
            #self.std_errors = np.array([float('nan') for i in range(self.nparams)])
        # Compute standard errors by inverting the Hessian
        inv_hessian = np.linalg.inv(hess(self.params))
        # Covariance matrix
        self.cov = 2. * inv_hessian
        # Calculate errors from diagonals of covariance matrix
        self.std_errors = np.sqrt(np.diag(self.cov))
        # Calculate fit values
        yfit = self.func(*self.params); 
        self.fit_data = self.data.copy()
        self.data['yfit'] = yfit
        # Calculate residuals, chi2, and reduced chi2
        residuals = (y-yfit)/yerr; self.residuals = residuals
        self.data['residuals'] = residuals
        self.chi2 = np.power(residuals,2).sum(); results['chi2'] = self.chi2
        self.nfree = self.ndata-self.nparams
        self.redchi2 = self.chi2/self.nfree; results['redchi2'] = self.redchi2
        # Scale errors by reduched chi-squared value
        if scale_covar:
            self.std_errors *= self.redchi2
        # Collect errors into results
        for i,p in enumerate(self.names):
            results['%s'%p] = self.params[i]
            results['%s_error'%p] = self.std_errors[i]
        # Calculate fit support and p-value
        ks = stats.kstest(residuals,'chi2',args=(self.nfree,))
        self.support,self.p = ks
        results['support'] = self.support
        results['p-value'] = self.p
        
        # Calculate loglikelihood
        self.loglikelihood = self.logL(self.params)
        # Calculate AIC, AICc, and BIC
        self.AIC = 2.*(self.nparams-self.loglikelihood); results['AIC'] = self.AIC
        self.AICc = self.AIC + (2*(pow(self.nparams,2)+self.nparams))/(self.ndata-self.nparams-1) ; results['AICc'] = self.AICc
        # Save results into private attribute
        self._results = results

    def mcmc(self, walkers = 32, steps = 2000, discard = 100, thin = 10):
        """
        Runs Monte Carlo sampler and computes standard errors as 0.5*(std_u - std_l)

        Parameters
        ----------
        walkers : int, optional
            Number of walkers, by default 32
        steps : int, optional
            Number of steps to take, by default 2000
        discard : int, optional
            Number of initial steps to discard, by default 100
        thin : int, optional
            Distance between independent steps, by default 10
        """
        MCMC.__init__(self,walkers=walkers,steps=steps)
        percentiles = self.calc_mc_errors(percentiles = [15.87,50,84.13],discard=discard,thin=thin)
        self.mcmc_params = np.zeros(self.nparams)
        self.mcmc_std_errors = np.zeros(self.nparams)
        for i,name in enumerate(self.names):
            std_l, median, std_u = percentiles[i]
            self.mcmc_params[i] = median
            self._results['%s_mcmc'%name] = median
            std_error = 0.5 * (std_u - std_l)
            self.mcmc_std_errors[i] = std_error
            self._results['%s_mcmc_error'%name] = 0.5 * (std_u - std_l)
    @property
    def results(self):
        """
        Dictionary of MLE fit results.

        Returns
        -------
        dict
            Dictionary of MLE fit results.
        """
        return self._results

class Gamma_Distribution:
    """Gamma probability distribution class.

    Parameters
    ----------
    shape : np.array
        Shape parameter.
    yhat : np.array
        Experimental values. 
    
    Attributes
    ----------
    shape : array-like
        Shape parameters.
    yhat : array-like
        Experimnetal values
    """
    def __init__(self,shape,yhat):
        self.yhat = yhat
        self.shape = shape
    def scale(self,ytrue):
        """ Calculate scale parameters.
        
        Parameters
        ----------
        ytrue : array-like
            True/theoretical values.

        Returns
        -------
        scale : array-like
            Scale parameters.
        """
        return ytrue/self.shape 
    def std(self,ytrue):
        """ Calculate standard deviation.

        Parameters
        ----------
        ytrue : array-like
            True/theoretical values.

        Returns
        -------
        std : array-like
            Standard deviations.
        """
        scale = self.scale(ytrue)
        return stats.gamma.std(self.shape, scale = scale)
    def pdf(self,ytrue):
        """ Calculate probability distribution function.

        Parameters
        ----------
        ytrue : array-like
            True/theoretical values.

        Returns
        -------
        pdf : callable
            Probability distribution function.
        """
        scale = self.scale(ytrue)
        return gamma.pdf(self.yhat/scale,self.shape)/scale
    def logpdf(self,ytrue):
        """ Log of the probability distribution function.

        Parameters
        ----------
        ytrue : array-like
            True/theoretical values.

        Returns
        -------
        logpdf : callable
            Log of the probability distribution function.
        """
        scale = self.scale(ytrue)
        return gamma.logpdf(self.yhat/scale,self.shape) - np.log(scale)
    def cdf(self,ytrue):
        """ Cumulative distribution function.

        Parameters
        ----------
        ytrue : array-like
            True/theoretical values.

        Returns
        -------
        cdf : callable
            Cumulative distribution function.
        """
        scale = self.scale(ytrue)
        return stats.gamma.cdf(self.yhat,self.shape,scale=scale)
    def logcdf(self, ytrue):
        """ Log of cumulative distribution function.

        Parameters
        ----------
        ytrue : array-like
            True/theoretical values.

        Returns
        -------
        logcdf : callable
            Log of cumulative distribution function.
        """
        scale = self.scale(ytrue)
        return gamma.logcdf(self.yhat,self.shape,scale = scale)
    def interval(self,ytrue,alpha = 0.95):
        """

        Parameters
        ----------
        ytrue : array-like
            True/theoretical values.
        alpha : float, optional
            Probability that a random variable will be drawn from the returned range. Each value should be in the range [0, 1]., by default 0.95

        Returns
        -------
        Interval : callable
            Endpoints of the range that contains fraction alpha [0, 1] of the distribution.
        """
        scale = self.scale(ytrue)
        return stats.gamma.interval(alpha,self.shape,scale = scale)